package org.xbib.gradle.plugin.jdeps.util

import org.gradle.api.Project
import org.gradle.api.Task
import org.gradle.api.internal.provider.Providers
import org.gradle.api.provider.Property
import org.gradle.api.provider.Provider
import org.gradle.api.provider.ProviderFactory

import static java.util.Objects.requireNonNull

final class SimpleBooleanState implements BooleanState {

    final Property<Boolean> property

    final Provider<Boolean> provider

    private final Project project

    SimpleBooleanState(Project project,
                       Property<Boolean> property,
                       Provider<Boolean> provider) {
        this.project = requireNonNull(project, "Argument 'project' must not be null")
        this.property = requireNonNull(property, "Argument 'property' must not be null")
        this.provider = requireNonNull(provider, "Argument 'provider' must not be null")
    }

    @Override
    boolean getValue() {
        booleanProvider(project.providers, property, provider, false).get()
    }

    static SimpleBooleanState of(Task task,
                                 String key,
                                 boolean defaultValue) {
        requireNonNull(task, "Argument 'task' must not be null")
        Project project = task.project
        Property<Boolean> property = project.objects.property(Boolean).convention(Providers.<Boolean>notDefined())
        Provider<Boolean> provider = booleanProvider(key, property, project, defaultValue)
        new SimpleBooleanState(project, property, provider)
    }

    static Provider<Boolean> booleanProvider(ProviderFactory providers,
                                             Property<Boolean> property,
                                             Provider<Boolean> provider,
                                             boolean defaultValue) {
        providers.provider {
            return provider ? provider.getOrElse(property.getOrElse(defaultValue)) : property.getOrElse(defaultValue)
        }
    }

    static Provider<Boolean> booleanProvider(String key,
                                             Property<Boolean> property,
                                             Project project,
                                             boolean defaultValue) {
        booleanProvider(toEnv(key), toProperty(key), property, project, defaultValue)
    }

    static Provider<Boolean> booleanProvider(String envKey,
                                             String propertyKey,
                                             Property<Boolean> property,
                                             Project project,
                                             boolean defaultValue) {
        project.providers.provider {
            String value = resolveValue(envKey, propertyKey, project)
            return !isBlank(value) ? Boolean.parseBoolean(value) : property.getOrElse(defaultValue)
        }
    }

    private static String resolveValue(String envKey, String propertyKey, Project project) {
        String value = System.getenv(envKey)
        if (isBlank(value)) {
            value = System.getProperty(propertyKey)
        }
        if (isBlank(value)) {
            value = project.findProperty(propertyKey) as String
        }
        return value
    }

    private static boolean isBlank(String str) {
        if (str == null || str.length() == 0) {
            return true
        }
        for (char c : str.toCharArray()) {
            if (!Character.isWhitespace(c)) {
                return false
            }
        }
        return true
    }

    private static String toEnv(String key) {
        key.toUpperCase().replace('.', '_')
    }

    private static String toProperty(String key) {
        key.uncapitalize().replace('_', '.')
    }
}
