package org.xbib.gradle.plugin.jdeps.util

import org.gradle.api.Project
import org.gradle.api.Task
import org.gradle.api.file.Directory
import org.gradle.api.file.DirectoryProperty
import org.gradle.api.internal.provider.Providers
import org.gradle.api.provider.Provider
import org.gradle.api.provider.ProviderFactory

import java.nio.file.Paths

import static java.util.Objects.requireNonNull

final class SimpleDirectoryState implements DirectoryState {

    final DirectoryProperty property

    final Provider<Directory> provider

    private final Project project

    SimpleDirectoryState(Project project,
                         DirectoryProperty property,
                         Provider<Directory> provider) {
        this.project = requireNonNull(project, "Argument 'project' must not be null")
        this.property = requireNonNull(property, "Argument 'property' must not be null")
        this.provider = requireNonNull(provider, "Argument 'provider' must not be null")
    }

    @Override
    Directory getValue() {
        directoryProvider(project.providers, property, provider, (Directory) null).get()
    }

    static SimpleDirectoryState of(Task task, String key, Directory defaultValue) {
        requireNonNull(task, "Argument 'task' must not be null")
        Project project = task.project
        DirectoryProperty property = project.objects.directoryProperty().convention(Providers.<Directory>notDefined())
        Provider<Directory> provider = directoryProvider(key, property, project, defaultValue)
        new SimpleDirectoryState(project, property, provider)
    }

    static Provider<Directory> directoryProvider(ProviderFactory providers,
                                                 DirectoryProperty property,
                                                 Provider<Directory> provider,
                                                 Directory defaultValue) {
        providers.provider {
            return provider ? provider.getOrElse(property.getOrElse(defaultValue)) : property.getOrElse(defaultValue)
        }
    }

    static Provider<Directory> directoryProvider(String key,
                                                 DirectoryProperty property,
                                                 Project project,
                                                 Directory defaultValue) {
        directoryProvider(toEnv(key), toProperty(key), property, project, defaultValue)
    }

    static Provider<Directory> directoryProvider(String envKey,
                                                 String propertyKey,
                                                 DirectoryProperty property,
                                                 Project project,
                                                 Directory defaultValue) {
        project.providers.provider {
            String value = resolveValue(envKey, propertyKey, project)
            if (!isBlank(value)) {
                DirectoryProperty p = project.objects.directoryProperty()
                p.set(Paths.get(value).toFile())
                return p.get()
            }
            return property.getOrElse(defaultValue)
        }
    }


    private static String resolveValue(String envKey, String propertyKey, Project project) {
        String value = System.getenv(envKey)
        if (isBlank(value)) {
            value = System.getProperty(propertyKey)
        }
        if (isBlank(value)) {
            value = project.findProperty(propertyKey) as String
        }
        return value
    }

    private static boolean isBlank(String str) {
        if (str == null || str.length() == 0) {
            return true
        }
        for (char c : str.toCharArray()) {
            if (!Character.isWhitespace(c)) {
                return false
            }
        }
        return true
    }

    private static String toEnv(String key) {
        key.toUpperCase().replace('.', '_')
    }

    private static String toProperty(String key) {
        key.uncapitalize().replace('_', '.')
    }
}
