package org.xbib.gradle.plugin.jdeps.util

import org.gradle.api.Project
import org.gradle.api.Task
import org.gradle.api.internal.provider.Providers
import org.gradle.api.provider.ListProperty
import org.gradle.api.provider.Provider
import org.gradle.api.provider.ProviderFactory

import static java.util.Objects.requireNonNull

final class SimpleListState implements ListState {

    final ListProperty<String> property

    final Provider<List<String>> provider

    private final Project project

    SimpleListState(Project project,
                    ListProperty<String> property,
                    Provider<List<String>> provider) {
        this.project = requireNonNull(project, "Argument 'project' must not be null")
        this.property = requireNonNull(property, "Argument 'property' must not be null")
        this.provider = requireNonNull(provider, "Argument 'provider' must not be null")
    }

    @Override
    List<String> getValue() {
        listProvider(project.providers, property, provider, Collections.<String> emptyList()).get()
    }

    static SimpleListState of(Task task, String key, List<String> defaultValue) {
        requireNonNull(task, "Argument 'task' must not be null")
        Project project = task.project
        ListProperty<String> property = project.objects.listProperty(String).convention(Providers.<List<String>>notDefined())
        Provider<List<String>> provider = listProvider(key, property, project, defaultValue)
        new SimpleListState(project, property, provider)
    }

    static Provider<List<String>> listProvider(ProviderFactory providers,
                                               ListProperty<String> property,
                                               Provider<List<String>> provider,
                                               List<String> defaultValue) {
        providers.provider {
            return provider ? provider.getOrElse(property.getOrElse(defaultValue)) : property.getOrElse(defaultValue)
        }
    }

    static Provider<List<String>> listProvider(String key,
                                               ListProperty<String> property,
                                               Project project,
                                               List<String> defaultValue) {
        listProvider(toEnv(key), toProperty(key), property, project, defaultValue)
    }

    static Provider<List<String>> listProvider(String envKey,
                                               String propertyKey,
                                               ListProperty<String> property,
                                               Project project,
                                               List<String> defaultValue) {
        project.providers.provider {
            if (property.present) {
                return property.get()
            }
            String value = resolveValue(envKey, propertyKey, project)
            if (!isBlank(value)) {
                List<String> list = new ArrayList<>()
                for (String v : value.split(',')) {
                    if (!isBlank(v)) list.add(v.trim())
                }
                return list
            }
            defaultValue
        }
    }

    private static String resolveValue(String envKey, String propertyKey, Project project) {
        String value = System.getenv(envKey)
        if (isBlank(value)) {
            value = System.getProperty(propertyKey)
        }
        if (isBlank(value)) {
            value = project.findProperty(propertyKey) as String
        }
        return value
    }

    private static boolean isBlank(String str) {
        if (str == null || str.length() == 0) {
            return true
        }
        for (char c : str.toCharArray()) {
            if (!Character.isWhitespace(c)) {
                return false
            }
        }
        return true
    }

    private static String toEnv(String key) {
        key.toUpperCase().replace('.', '_')
    }

    private static String toProperty(String key) {
        key.uncapitalize().replace('_', '.')
    }
}
