package org.xbib.gradle.plugin.jdeps.util

import org.gradle.api.Project
import org.gradle.api.Task
import org.gradle.api.internal.provider.Providers
import org.gradle.api.provider.MapProperty
import org.gradle.api.provider.Provider
import org.gradle.api.provider.ProviderFactory

import static java.util.Objects.requireNonNull

final class SimpleMapState implements MapState {

    final MapProperty<String, String> property

    final Provider<Map<String, String>> provider

    private final Project project

    SimpleMapState(Project project,
                   MapProperty<String, String> property,
                   Provider<Map<String, String>> provider) {
        this.project = requireNonNull(project, "Argument 'project' must not be null")
        this.property = requireNonNull(property, "Argument 'property' must not be null")
        this.provider = requireNonNull(provider, "Argument 'provider' must not be null")
    }

    @Override
    Map<String, String> getValue() {
        mapProvider(project.providers, property, provider, Collections.<String, String> emptyMap()).get()
    }

    static SimpleMapState of(Task task, String key, Map<String, String> defaultValue) {
        requireNonNull(task, "Argument 'task' must not be null")
        Project project = task.project
        MapProperty<String, String> property = project.objects.mapProperty(String, String).convention(Providers.<Map<String, String>>notDefined())
        Provider<Map<String, String>> provider = mapProvider(key, property, project, defaultValue)
        new SimpleMapState(project, property, provider)
    }

    static <K, V> Provider<Map<K, V>> mapProvider(ProviderFactory providers,
                                                  MapProperty<K, V> property,
                                                  Provider<Map<K, V>> provider,
                                                  Map<K, V> defaultValue) {
        providers.provider {
            Map<K, V> map = new LinkedHashMap<>(defaultValue)
            if (property.present) {
                map.putAll(property.get())
            }
            if (provider) {
                map.putAll(provider.get())
            }
            return map
        }
    }

    static <K,V> Provider<Map<K, V>> mapProvider(String key,
                                                 MapProperty<K, V> property,
                                                 Project project,
                                                 Map<K, V> defaultValue) {
        mapProvider(toEnv(key), toProperty(key), property, project, defaultValue)
    }

    static <K, V> Provider<Map<K, V>> mapProvider(String envKey,
                                                  String propertyKey,
                                                  MapProperty<K, V> property,
                                                  Project project,
                                                  Map<K, V> defaultValue) {
        project.providers.provider {
            String value = resolveValue(envKey, propertyKey, project)
            if (i!sBlank(value)) {
                Map<K, V> map = new LinkedHashMap<>()
                for (String val : value.split(',')) {
                    String[] kv = val.split('=')
                    if (kv.length == 2) {
                        map.put(kv[0] as K, kv[1] as V)
                    }
                }
                return map
            }
            return property.orElse(defaultValue).get()
        }
    }

    private static String resolveValue(String envKey, String propertyKey, Project project) {
        String value = System.getenv(envKey)
        if (isBlank(value)) {
            value = System.getProperty(propertyKey)
        }
        if (isBlank(value)) {
            value = project.findProperty(propertyKey) as String
        }
        return value
    }

    private static boolean isBlank(String str) {
        if (str == null || str.length() == 0) {
            return true
        }
        for (char c : str.toCharArray()) {
            if (!Character.isWhitespace(c)) {
                return false
            }
        }
        return true
    }

    private static String toEnv(String key) {
        key.toUpperCase().replace('.', '_')
    }

    private static String toProperty(String key) {
        key.uncapitalize().replace('_', '.')
    }
}
