package org.xbib.gradle.plugin.jdeps.util

import org.gradle.api.Project
import org.gradle.api.Task
import org.gradle.api.internal.provider.Providers
import org.gradle.api.provider.Property
import org.gradle.api.provider.Provider
import org.gradle.api.provider.ProviderFactory

import static java.util.Objects.requireNonNull

final class SimpleStringState implements StringState {

    final Property<String> property

    final Provider<String> provider

    private final Project project

    @Override
    String getValue() {
        stringProvider(project.providers, property, provider, '').get()
    }

    SimpleStringState(Project project,
                      Property<String> property,
                      Provider<String> provider) {
        this.project = requireNonNull(project,  "Argument 'project' must not be null")
        this.property = requireNonNull(property, "Argument 'property' must not be null")
        this.provider = requireNonNull(provider, "Argument 'provider' must not be null")
    }

    static SimpleStringState of(Task task, String key, String defaultValue) {
        requireNonNull(task, "Argument 'task' must not be null")
        Project project = task.project
        Property<String> property = project.objects.property(String).convention(Providers.<String>notDefined())
        Provider<String> provider = stringProvider(key, property, project, defaultValue)
        new SimpleStringState(project, property, provider)
    }

    static Provider<String> stringProvider(ProviderFactory providers,
                                           Property<String> property,
                                           Provider<String> provider,
                                           String defaultValue) {
        providers.provider {
            return provider ? provider.getOrElse(property.getOrElse(defaultValue)) : property.getOrElse(defaultValue)
        }
    }

    static Provider<String> stringProvider(String key,
                                           Property<String> property,
                                           Project project,
                                           String defaultValue) {
        stringProvider(toEnv(key), toProperty(key), property, project, defaultValue)
    }

    static Provider<String> stringProvider(String envKey,
                                           String propertyKey,
                                           Property<String> property,
                                           Project project,
                                           String defaultValue) {
        project.providers.provider {
            String value = resolveValue(envKey, propertyKey, project)
            return !isBlank(value) ? value : property.getOrElse(defaultValue)
        }
    }

    private static String resolveValue(String envKey, String propertyKey, Project project) {
        String value = System.getenv(envKey)
        if (isBlank(value)) {
            value = System.getProperty(propertyKey)
        }
        if (isBlank(value)) {
            value = project.findProperty(propertyKey) as String
        }
        return value
    }

    private static boolean isBlank(String str) {
        if (str == null || str.length() == 0) {
            return true
        }
        for (char c : str.toCharArray()) {
            if (!Character.isWhitespace(c)) {
                return false
            }
        }
        return true
    }

    private static String toEnv(String key) {
        key.toUpperCase().replace('.', '_')
    }

    private static String toProperty(String key) {
        key.uncapitalize().replace('_', '.')
    }
}
