package org.xbib.gradle.plugin.jpackage;

import java.util.Arrays;
import java.util.Collections;
import javax.inject.Inject;
import org.gradle.api.Action;
import org.gradle.api.NamedDomainObjectContainer;
import org.gradle.api.NonNullApi;
import org.gradle.api.Project;
import org.gradle.api.Task;
import org.gradle.api.artifacts.Configuration;
import org.gradle.api.artifacts.ConfigurationContainer;
import org.gradle.api.attributes.Attribute;
import org.gradle.api.attributes.Bundling;
import org.gradle.api.attributes.Category;
import org.gradle.api.attributes.LibraryElements;
import org.gradle.api.attributes.Usage;
import org.gradle.api.attributes.java.TargetJvmEnvironment;
import org.gradle.api.file.ConfigurableFileCollection;
import org.gradle.api.file.DirectoryProperty;
import org.gradle.api.model.ObjectFactory;
import org.gradle.api.plugins.ApplicationPlugin;
import org.gradle.api.plugins.JavaApplication;
import org.gradle.api.plugins.JavaPluginExtension;
import org.gradle.api.provider.Property;
import org.gradle.api.tasks.JavaExec;
import org.gradle.api.tasks.SourceSet;
import org.gradle.api.tasks.SourceSetContainer;
import org.gradle.api.tasks.TaskContainer;
import org.gradle.api.tasks.TaskProvider;
import org.gradle.jvm.toolchain.JavaToolchainService;
import org.gradle.nativeplatform.MachineArchitecture;
import org.gradle.nativeplatform.OperatingSystemFamily;
import static org.gradle.language.base.plugins.LifecycleBasePlugin.ASSEMBLE_TASK_NAME;
import static org.gradle.language.base.plugins.LifecycleBasePlugin.BUILD_GROUP;
import static org.gradle.nativeplatform.MachineArchitecture.ARCHITECTURE_ATTRIBUTE;
import static org.gradle.nativeplatform.OperatingSystemFamily.LINUX;
import static org.gradle.nativeplatform.OperatingSystemFamily.MACOS;
import static org.gradle.nativeplatform.OperatingSystemFamily.OPERATING_SYSTEM_ATTRIBUTE;
import static org.gradle.nativeplatform.OperatingSystemFamily.WINDOWS;

@NonNullApi
public abstract class JPackageExtension {

    public abstract Property<String> getApplicationName();

    public abstract Property<String> getApplicationVersion();

    public abstract Property<String> getApplicationDescription();

    public abstract Property<String> getVendor();

    public abstract Property<String> getCopyright();

    public abstract DirectoryProperty getJpackageResources();

    public abstract ConfigurableFileCollection getResources();

    private final NamedDomainObjectContainer<JPackageTarget> targets = getObjects().domainObjectContainer(JPackageTarget.class);

    @Inject
    protected abstract JavaToolchainService getJavaToolchains();

    @Inject
    protected abstract ObjectFactory getObjects();

    @Inject
    protected abstract Project getProject();

    public JPackageTarget target(String label, Action<? super JPackageTarget> action) {
        JPackageTarget target;
        if (targets.getNames().contains(label)) {
            target = targets.getByName(label);
        } else {
            target = targets.create(label, this::newTarget);
        }
        action.execute(target);
        return target;
    }

    public JPackageTarget primaryTarget(JPackageTarget target) {
        SourceSetContainer sourceSets = getProject().getExtensions().getByType(SourceSetContainer.class);
        ConfigurationContainer configurations = getProject().getConfigurations();
        sourceSets.all(sourceSet -> {
            configureTargetAttributes(configurations.getByName(sourceSet.getCompileClasspathConfigurationName()), target);
            configureTargetAttributes(configurations.getByName(sourceSet.getRuntimeClasspathConfigurationName()), target);
            configurations.matching(conf -> "mainRuntimeClasspath".equals(conf.getName())).all(conf -> configureTargetAttributes(conf, target));
        });

        return target;
    }

    private void newTarget(JPackageTarget target) {
        target.getPackageTypes().convention(target.getOperatingSystem().map(os -> switch (os) {
            case WINDOWS -> Arrays.asList("exe", "msi");
            case MACOS -> Arrays.asList("pkg", "dmg");
            case LINUX -> Arrays.asList("rpm");
            default -> Collections.emptyList();
        }));
        ConfigurationContainer configurations = getProject().getConfigurations();
        SourceSetContainer sourceSets = getProject().getExtensions().getByType(SourceSetContainer.class);
        sourceSets.all(sourceSet -> {
            Configuration internal = maybeCreateInternalConfiguration();
            configurations.create(target.getName() + capitalize(sourceSet.getCompileClasspathConfigurationName()), c -> {
                c.setCanBeConsumed(false);
                c.setVisible(false);
                configureJavaStandardAttributes(c, Usage.JAVA_API);
                configureTargetAttributes(c, target);
                c.extendsFrom(
                        configurations.getByName(sourceSet.getImplementationConfigurationName()),
                        configurations.getByName(sourceSet.getCompileOnlyConfigurationName()),
                        internal
                );
            });
            Configuration runtimeClasspath = configurations.create(target.getName() + capitalize(sourceSet.getRuntimeClasspathConfigurationName()), c -> {
                c.setCanBeConsumed(false);
                c.setVisible(false);
                configureJavaStandardAttributes(c, Usage.JAVA_RUNTIME);
                configureTargetAttributes(c, target);
                c.extendsFrom(
                        configurations.getByName(sourceSet.getImplementationConfigurationName()),
                        configurations.getByName(sourceSet.getRuntimeOnlyConfigurationName()),
                        internal
                );
            });
            if (SourceSet.isMain(sourceSet)) {
                getProject().getPlugins().withType(ApplicationPlugin.class, p -> registerTargetSpecificTasks(target, sourceSet.getJarTaskName(), runtimeClasspath));
            }
        });
    }

    private void configureJavaStandardAttributes(Configuration resolvable, String usage) {
        resolvable.getAttributes().attribute(Usage.USAGE_ATTRIBUTE, getObjects().named(Usage.class, usage));
        resolvable.getAttributes().attribute(Category.CATEGORY_ATTRIBUTE, getObjects().named(Category.class, Category.LIBRARY));
        resolvable.getAttributes().attribute(LibraryElements.LIBRARY_ELEMENTS_ATTRIBUTE, getObjects().named(LibraryElements.class, LibraryElements.JAR));
        resolvable.getAttributes().attribute(TargetJvmEnvironment.TARGET_JVM_ENVIRONMENT_ATTRIBUTE, getObjects().named(TargetJvmEnvironment.class, TargetJvmEnvironment.STANDARD_JVM));
        resolvable.getAttributes().attribute(Bundling.BUNDLING_ATTRIBUTE, getObjects().named(Bundling.class, Bundling.EXTERNAL));
        resolvable.getAttributes().attribute(Attribute.of("javaModule", Boolean.class), true);
    }

    private void configureTargetAttributes(Configuration resolvable, JPackageTarget target) {
        resolvable.getAttributes().attributeProvider(OPERATING_SYSTEM_ATTRIBUTE, target.getOperatingSystem().map(name -> getObjects().named(OperatingSystemFamily.class, name)));
        resolvable.getAttributes().attributeProvider(ARCHITECTURE_ATTRIBUTE, target.getArchitecture().map(name -> getObjects().named(MachineArchitecture.class, name)));
    }

    private void registerTargetSpecificTasks(JPackageTarget target, String applicationJarTask, Configuration runtimeClasspath) {
        TaskContainer tasks = getProject().getTasks();
        JavaPluginExtension java = getProject().getExtensions().getByType(JavaPluginExtension.class);
        JavaApplication application = getProject().getExtensions().getByType(JavaApplication.class);
        TaskProvider<JPackageTask> jpackage = tasks.register("jpackage" + capitalize(target.getName()), JPackageTask.class, t -> {
            t.getJavaInstallation().convention(getJavaToolchains().compilerFor(java.getToolchain()).get().getMetadata());
            t.getOperatingSystem().convention(target.getOperatingSystem());
            t.getArchitecture().convention(target.getArchitecture());
            t.getMainModule().convention(application.getMainModule());
            t.getVersion().convention(getApplicationVersion());
            t.getModulePath().from(tasks.named(applicationJarTask));
            t.getModulePath().from(runtimeClasspath);
            t.getApplicationName().convention(getApplicationName());
            t.getJpackageResources().convention(getJpackageResources().dir(target.getOperatingSystem()));
            t.getApplicationDescription().convention(getApplicationDescription());
            t.getVendor().convention(getVendor());
            t.getCopyright().convention(getCopyright());
            t.getJavaOptions().convention(application.getApplicationDefaultJvmArgs());
            t.getOptions().convention(target.getOptions());
            t.getPackageTypes().convention(target.getPackageTypes());
            t.getResources().from(getResources());
            t.getDestination().convention(getProject().getLayout().getBuildDirectory().dir("packages/" + target.getName()));
            t.getTempDirectory().convention(getProject().getLayout().getBuildDirectory().dir("tmp/jpackage/" + target.getName()));
        });
        tasks.register("run" + capitalize(target.getName()), JavaExec.class, t -> {
            t.setGroup(ApplicationPlugin.APPLICATION_GROUP);
            t.setDescription("Run this project as a JVM application on " + target.getName());
            t.getJavaLauncher().convention(getJavaToolchains().launcherFor(java.getToolchain()));
            t.getMainModule().convention(application.getMainModule());
            t.getMainClass().convention(application.getMainClass());
            t.setJvmArgs(application.getApplicationDefaultJvmArgs());
            t.classpath(tasks.named("jar"), runtimeClasspath);
        });
        String targetAssembleLifecycle = "assemble" + capitalize(target.getName());
        if (!tasks.getNames().contains(targetAssembleLifecycle)) {
            TaskProvider<Task> lifecycleTask = tasks.register(targetAssembleLifecycle, t -> {
                t.setGroup(BUILD_GROUP);
                t.setDescription("Builds this project for " + target.getName());
            });
            tasks.named(ASSEMBLE_TASK_NAME, t -> t.dependsOn(lifecycleTask));
        }
        tasks.named(targetAssembleLifecycle, t -> t.dependsOn(jpackage));
    }

    private Configuration maybeCreateInternalConfiguration() {
        Configuration internal = getProject().getConfigurations().findByName("internal");
        if (internal != null) {
            return internal;
        }
        return getProject().getConfigurations().create("internal", i -> {
            i.setCanBeResolved(false);
            i.setCanBeConsumed(false);
        });
    }

    private String capitalize(String s) {
        return s.substring(0, 1).toUpperCase() + s.substring(1);
    }
}
