package org.xbib.gradle.plugin.jpackage;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Arrays;
import java.util.List;
import javax.inject.Inject;
import org.gradle.api.DefaultTask;
import org.gradle.api.file.ConfigurableFileCollection;
import org.gradle.api.file.Directory;
import org.gradle.api.file.DirectoryProperty;
import org.gradle.api.internal.file.FileOperations;
import org.gradle.api.provider.ListProperty;
import org.gradle.api.provider.Property;
import org.gradle.api.tasks.CacheableTask;
import org.gradle.api.tasks.Classpath;
import org.gradle.api.tasks.Input;
import org.gradle.api.tasks.InputDirectory;
import org.gradle.api.tasks.InputFiles;
import org.gradle.api.tasks.Internal;
import org.gradle.api.tasks.Nested;
import org.gradle.api.tasks.Optional;
import org.gradle.api.tasks.OutputDirectory;
import org.gradle.api.tasks.PathSensitive;
import org.gradle.api.tasks.PathSensitivity;
import org.gradle.api.tasks.TaskAction;
import org.gradle.jvm.toolchain.JavaInstallationMetadata;
import org.gradle.process.ExecOperations;
import static java.util.Objects.requireNonNull;
import static org.gradle.nativeplatform.OperatingSystemFamily.WINDOWS;

@CacheableTask
public abstract class JPackageTask extends DefaultTask {

    @Nested
    public abstract Property<JavaInstallationMetadata> getJavaInstallation();

    @Input
    public abstract Property<String> getOperatingSystem();

    @Input
    public abstract Property<String> getArchitecture();

    @Input
    public abstract Property<String> getMainModule();

    @Input
    public abstract Property<String> getVersion();

    @Classpath
    public abstract ConfigurableFileCollection getModulePath();

    @Input
    public abstract Property<String> getApplicationName();

    @Input
    @Optional
    public abstract Property<String> getApplicationDescription();

    @InputDirectory
    @PathSensitive(PathSensitivity.RELATIVE)
    public abstract DirectoryProperty getJpackageResources();

    @InputFiles
    @PathSensitive(PathSensitivity.RELATIVE)
    public abstract ConfigurableFileCollection getResources();

    @Input
    @Optional
    public abstract Property<String> getVendor();

    @Input
    @Optional
    public abstract Property<String> getCopyright();

    @Input
    public abstract ListProperty<String> getJavaOptions();

    @Input
    public abstract ListProperty<String> getOptions();

    @Input
    public abstract ListProperty<String> getPackageTypes();

    @OutputDirectory
    public abstract DirectoryProperty getDestination();

    @Internal
    public abstract DirectoryProperty getTempDirectory();

    @Inject
    protected abstract FileOperations getFiles();

    @Inject
    protected abstract ExecOperations getExec();

    @TaskAction
    public void runJpackage() throws Exception {
        getFiles().delete(getTempDirectory());
        getFiles().delete(getDestination());
        String os = getOperatingSystem().get();
        String arch = getArchitecture().get();
        String hostOs = System.getProperty("os.name").replace(" ", "").toLowerCase();
        String hostArch = System.getProperty("os.arch");
        validateHostSystem(arch, hostArch, os, hostOs);
        Directory resourcesDir = getTempDirectory().get().dir("jpackage-resources");
        Directory appImageParent = getTempDirectory().get().dir("app-image");
        resourcesDir.getAsFile().mkdirs();
        getFiles().copy(c -> {
            c.from(getJpackageResources());
            c.into(resourcesDir);
            c.rename(f -> f.replace("icon", getApplicationName().get()));
        });
        String executableName = WINDOWS.equals(os) ? "jpackage.exe" : "jpackage";
        String jpackage = getJavaInstallation().get()
                .getInstallationPath().file("bin/" + executableName)
                .getAsFile().getAbsolutePath();
        getExec().exec(e -> {
            e.commandLine(
                    jpackage,
                    "--type",
                    "app-image",
                    "--module",
                    getMainModule().get(),
                    "--resource-dir",
                    resourcesDir.getAsFile().getPath(),
                    "--app-version",
                    getVersion().get(),
                    "--module-path",
                    getModulePath().getAsPath(),
                    "--name",
                    getApplicationName().get(),
                    "--dest",
                    appImageParent.getAsFile().getPath()
            );
            if (getApplicationDescription().isPresent()) {
                e.args("--description", getApplicationDescription().get());
            }
            if (getVendor().isPresent()) {
                e.args("--vendor", getVendor().get());
            }
            if (getCopyright().isPresent()) {
                e.args("--copyright", getCopyright().get());
            }
            for (String javaOption : getJavaOptions().get()) {
                e.args("--java-options", javaOption);
            }
        });
        File appImageFolder = requireNonNull(appImageParent.getAsFile().listFiles())[0];
        File appResourcesFolder;
        if (os.contains("macos")) {
            appResourcesFolder = new File(appImageFolder, "Contents/app");
        } else if (os.contains("windows")) {
            appResourcesFolder = new File(appImageFolder, "app");
        } else {
            appResourcesFolder = new File(appImageFolder, "lib/app");
        }
        getFiles().copy(c -> {
            c.from(getResources());
            c.into(appResourcesFolder);
        });
        getPackageTypes().get().forEach(packageType ->
                getExec().exec(e -> {
                    e.commandLine(
                            jpackage,
                            "--type",
                            packageType,
                            "--app-image",
                            appImageFolder.getPath(),
                            "--dest",
                            getDestination().get().getAsFile().getPath()
                    );
                    for (String option : getOptions().get()) {
                        e.args(option);
                    }
                })
        );
        generateChecksums();
    }

    private void generateChecksums() throws NoSuchAlgorithmException, IOException {
        File destination = getDestination().get().getAsFile();
        List<File> allFiles = Arrays.stream(requireNonNull(destination.listFiles()))
                .filter(File::isFile)
                .toList();
        for (File result : allFiles) {
            MessageDigest digest = MessageDigest.getInstance("SHA-256");
            byte[] encoded = digest.digest(Files.readAllBytes(result.toPath()));
            Files.write(new File(destination, result.getName() + ".sha256").toPath(), bytesToHex(encoded).getBytes());
        }
    }

    private String bytesToHex(byte[] hash) {
        StringBuilder hexString = new StringBuilder(2 * hash.length);
        for (byte b : hash) {
            String hex = Integer.toHexString(0xff & b);
            if (hex.length() == 1) {
                hexString.append('0');
            }
            hexString.append(hex);
        }
        return hexString.toString();
    }

    private void validateHostSystem(String arch, String hostArch, String os, String hostOs) {
        if (os.contains("macos")) {
            if (!hostOs.contains(os)) {
                wrongHostSystemError(hostOs, os);
            }
        } else if (os.contains("windows")) {
            if (!hostOs.contains(os)) {
                wrongHostSystemError(hostOs, os);
            }
        } else {
            if (hostOs.contains("windows") || hostOs.contains("macos")) {
                wrongHostSystemError(hostOs, os);
            }
        }
        if (arch.contains("64") && !hostArch.contains("64")) {
            wrongHostSystemError(hostArch, arch);
        }
        if (arch.contains("aarch") && !hostArch.contains("aarch")) {
            wrongHostSystemError(hostArch, arch);
        }
        if (!arch.contains("aarch") && hostArch.contains("aarch")) {
            wrongHostSystemError(hostArch, arch);
        }
    }

    private void wrongHostSystemError(String hostOs, String os) {
        throw new RuntimeException("Running on " + hostOs + "; cannot build for " + os);
    }
}
